/* No representations are made concerning either the merchantability of
   this software or the suitability of this software for any particular
   purpose. It is provided "as is" without express or implied warranty
   of any kind.

   License to copy and use this software is granted provided that these
   notices are retained in any copies of any part of this documentation
   and/or software.
 */

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include "config.h"

MicrolazeConfig::MicrolazeConfig ()
{
	_wavelength = 2;
	_energyRange = 2;
	_energyLevel = 1000;
	_pulseRate = 40;
	_markerSizeX = 100;
	_markerSizeY = 100;
	_scoreStepSize = .15;
	_cutStepSize = .025;
	_distanceToRefocus = .5;
	_focusOffcenterDistance = .1;
	_scoreHoldTime = 2.;
	_cutHoldTime = 1.;
	_minFocusScore = 800;
	_cutFeedSpeed = 100;
}

// Return TRUE for success, FALSE if trouble opening file.
// Print warnings for errors in config file.
BOOL MicrolazeConfig::loadFromFile (char *filename)
{
	char buffer[80], *value;
	int line, i;
	
	FILE *infile = fopen (filename, "r");
	if (infile == NULL) {
		printf ("Cannot open config file: %s\n", filename);
		return FALSE;
	}

	line = 0;
	while (1) {
		if (fgets (buffer, sizeof (buffer), infile) == NULL)
			break;
		++line;

		// Remove line terminators and spaces from end of string
		for (i = strlen (buffer) - 1; i >= 0; --i) {
			if (buffer[i] == '\012' || buffer[i] == '\015' || buffer[i] == ' ')
				buffer[i] = '\0';
			else
				break;
		}

		value = strchr (buffer, '=');
		if (value == NULL)
			// No '=' on line
			continue;

		// replace '=' with null termination so the parameter type is a normal string
		value[0] = '\0';
		// point value past the '='
		++value;

		if (stricmp (buffer, wavelengthName()) == 0)
			setWavelength (value);
		else if (stricmp (buffer, energyRangeName()) == 0)
			setEnergyRange (value);
		else if (stricmp (buffer, energyLevelName()) == 0)
			setEnergyLevel (value);
		else if (stricmp (buffer, pulseRateName()) == 0)
			setPulseRate (value);
		else if (stricmp (buffer, markerSizeXName()) == 0)
			setMarkerSizeX (value);
		else if (stricmp (buffer, markerSizeYName()) == 0)
			setMarkerSizeY (value);
		else if (stricmp (buffer, scoreStepSizeName()) == 0)
			setScoreStepSize (value);
		else if (stricmp (buffer, cutStepSizeName()) == 0)
			setCutStepSize (value);
		else if (stricmp (buffer, distanceToRefocusName()) == 0)
			setDistanceToRefocus (value);
		else if (stricmp (buffer, focusOffcenterDistanceName()) == 0)
			setFocusOffcenterDistance (value);
		else if (stricmp (buffer, scoreHoldTimeName()) == 0)
			setScoreHoldTime (value);
		else if (stricmp (buffer, cutHoldTimeName()) == 0)
			setCutHoldTime (value);
		else if (stricmp (buffer, minFocusScoreName()) == 0)
			setMinFocusScore (value);
		else if (stricmp (buffer, cutFeedSpeedName()) == 0)
			setCutFeedSpeed (value);
		else
			printf ("WARNING: config line %d: unrecognized parameter %s\n", line, buffer);
	}

	fclose (infile);
	return TRUE;
}

// Return TRUE for success, FALSE if trouble opening file.
BOOL MicrolazeConfig::saveToFile (char *filename)
{
	FILE *outfile;

	outfile = fopen (filename, "w");
	if (outfile == NULL) {
		printf ("Cannot open config file for write: %s\n", filename);
		return FALSE;
	}

	fprintf (outfile, "%s=%s\n", wavelengthName(), wavelengthString());
	fprintf (outfile, "%s=%s\n", energyRangeName(), energyRangeString());
	fprintf (outfile, "%s=%d\n", energyLevelName(), energyLevel());
	fprintf (outfile, "%s=%d\n", pulseRateName(), pulseRate());
	fprintf (outfile, "%s=%d\n", markerSizeXName(), markerSizeX());
	fprintf (outfile, "%s=%d\n", markerSizeYName(), markerSizeY());
	fprintf (outfile, "%s=%lf\n", scoreStepSizeName(), scoreStepSize());
	fprintf (outfile, "%s=%lf\n", cutStepSizeName(), cutStepSize());
	fprintf (outfile, "%s=%lf\n", distanceToRefocusName(), distanceToRefocus());
	fprintf (outfile, "%s=%lf\n", focusOffcenterDistanceName(), focusOffcenterDistance());
	fprintf (outfile, "%s=%lf\n", scoreHoldTimeName(), scoreHoldTime());
	fprintf (outfile, "%s=%lf\n", cutHoldTimeName(), cutHoldTime());
	fprintf (outfile, "%s=%d\n", minFocusScoreName(), minFocusScore());
	fprintf (outfile, "%s=%d\n", cutFeedSpeedName(), cutFeedSpeed());

	fclose (outfile);
	return TRUE;
}

char *MicrolazeConfig::wavelengthString()
{
	if (_wavelength == 1)
		return ("IR");
	else if (_wavelength == 2)
		return ("Green");
	else
		return ("UV");
}

// Prints warning if illegal value
void MicrolazeConfig::setWavelength (char *value)
{
	if (stricmp (value, "ir") == 0)
		_wavelength = 1;
	else if (stricmp (value, "green") == 0)
		_wavelength = 2;
	else if (stricmp (value, "uv") == 0)
		_wavelength = 3;
	else
		printf ("WARNING: config: wavelength must be IR, Green or UV. Got: %s\n", value);
}

char *MicrolazeConfig::energyRangeString()
{
	if (_energyRange == 1)
		return ("low");
	else
		return ("high");
}

// Prints warning if illegal value
void MicrolazeConfig::setEnergyRange (char *value)
{
	if (stricmp (value, "low") == 0)
		_energyRange = 1;
	else if (stricmp (value, "high") == 0)
		_energyRange = 2;
	else
		printf ("WARNING: config: energyrange must be low or high. Got: %s\n", value);
}

// Reads an integer from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setEnergyLevel (char *value)
{
	int intvalue;

	if (sscanf (value, "%d", &intvalue) != 1 ||
		intvalue < 0 || intvalue > 1000)
		printf ("WARNING: config: energylevel must be from 0 to 1000. Got: %s\n", value);
	else
		_energyLevel = intvalue;
}

// Reads an integer from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setPulseRate (char *value)
{
	int intvalue;

	if (sscanf (value, "%d", &intvalue) != 1 ||
		intvalue < 1 || intvalue > 40)
		printf ("WARNING: config: pulserate must be from 1 to 40. Got: %s\n", value);
	else
		_pulseRate = intvalue;
}

// Reads an integer from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setMarkerSizeX (char *value)
{
	int intvalue;

	if (sscanf (value, "%d", &intvalue) != 1 ||
		intvalue < 0 || intvalue > 100)
		printf ("WARNING: config: markersizeX must be from 0 to 100. Got: %s\n", value);
	else
		_markerSizeX = intvalue;
}

// Reads an integer from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setMarkerSizeY (char *value)
{
	int intvalue;

	if (sscanf (value, "%d", &intvalue) != 1 ||
		intvalue < 0 || intvalue > 100)
		printf ("WARNING: config: markersizeY must be from 0 to 100. Got: %s\n", value);
	else
		_markerSizeY = intvalue;
}

// Reads a double from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setScoreStepSize (char *value)
{
	double doublevalue;

	if (sscanf (value, "%lf", &doublevalue) != 1 ||	doublevalue <= 0.)
		printf ("WARNING: config: scorestepsize must be > 0. Got: %s\n", value);
	else
		_scoreStepSize = doublevalue;
}

// Reads a double from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setCutStepSize (char *value)
{
	double doublevalue;

	if (sscanf (value, "%lf", &doublevalue) != 1 ||	doublevalue <= 0.)
		printf ("WARNING: config: cutstepsize must be > 0. Got: %s\n", value);
	else
		_cutStepSize = doublevalue;
}

// Reads a double from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setDistanceToRefocus (char *value)
{
	double doublevalue;

	if (sscanf (value, "%lf", &doublevalue) != 1 ||	doublevalue < 0.)
		printf ("WARNING: config: distanceToRefocus must be >= 0. Got: %s\n", value);
	else
		_distanceToRefocus = doublevalue;
}

// Reads a double from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setFocusOffcenterDistance (char *value)
{
	double doublevalue;

	if (sscanf (value, "%lf", &doublevalue) != 1 ||	doublevalue < 0.)
		printf ("WARNING: config: focusOffcenterDistance must be >= 0. Got: %s\n", value);
	else
		_focusOffcenterDistance = doublevalue;
}

// Reads a double from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setScoreHoldTime (char *value)
{
	double doublevalue;

	if (sscanf (value, "%lf", &doublevalue) != 1 ||	doublevalue < 0.)
		printf ("WARNING: config: scoreHoldTime must be >= 0. Got: %s\n", value);
	else
		_scoreHoldTime = doublevalue;
}

// Reads a double from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setCutHoldTime (char *value)
{
	double doublevalue;

	if (sscanf (value, "%lf", &doublevalue) != 1 ||	doublevalue < 0.)
		printf ("WARNING: config: cutHoldTime must be >= 0. Got: %s\n", value);
	else
		_cutHoldTime = doublevalue;
}

// Reads an integer from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setMinFocusScore (char *value)
{
	int intvalue;

	if (sscanf (value, "%d", &intvalue) != 1 || intvalue < 0)
		printf ("WARNING: config: minFocusScore must be >= 0. Got: %s\n", value);
	else
		_minFocusScore = intvalue;
}

// Reads an integer from the value string.
// Prints warning if illegal value
void MicrolazeConfig::setCutFeedSpeed (char *value)
{
	int intvalue;

	if (sscanf (value, "%d", &intvalue) != 1 || intvalue < 1 || intvalue > 100)
		printf ("WARNING: config: cutFeedSpeed must be from 1 to 100. Got: %s\n", value);
	else
		_cutFeedSpeed = intvalue;
}

