/* No representations are made concerning either the merchantability of
   this software or the suitability of this software for any particular
   purpose. It is provided "as is" without express or implied warranty
   of any kind.

   License to copy and use this software is granted provided that these
   notices are retained in any copies of any part of this documentation
   and/or software.
 */

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <windows.h>
#include <windowsx.h>
#include <process.h>
#include <math.h>
#include <time.h>
#include "serial.h"
#include "quiklaze.h"
#include "prior.h"
#include "config.h"
#include "offset.h"

// Make a console version of Prior which gets the interrupt from the keyboard.
// This is in contrast to a GUI which might use a Cancel button.
class ConsolePrior : public Prior {
public:
	virtual BOOL checkInterrupt();
	virtual BOOL doInterrupt();
};

// Return FALSE to abort, TRUE to continue operation.
// This may stop firing.  Calling routine must do startFiring() afterwards if necessary.
// Override default in base class.
BOOL ConsolePrior::checkInterrupt () 
{
	if (!_kbhit())
		return (TRUE);
	
	return (doInterrupt ());
}

// Return FALSE to abort, TRUE to continue operation.
// This will stop firing.  Calling routine must do startFiring() afterwards if necessary.
// Override default in base class.
BOOL ConsolePrior::doInterrupt () 
{
	int c;

	stopFiring();

	while (1) {
		// clear the input buffer
		while (_kbhit())
			c = _getch();
		printf ("                                                                           \r");
		puts ("INTERRUPTED!  Hit q to quit, c to continue...");

		// get one keystroke
		while (!_kbhit());
		c = _getch();

		if (c == 'q' || c == 'Q')
			return (FALSE);
		if (c == 'c' || c == 'C')
			return (TRUE);
	}
}

void Usage() {
	puts ("usage: microlaze [-c configfile] [-o offsetfile] [-l startline]");
	puts ("  [-ox xOffset] [-oy yOffset] [-oa angleOffset] [-s scorefile] [cutfile]");
	puts (" -c: configuration file produced by 'microlazeconfig configfile'");
	puts (" -o: offset file produced by 'microlazeoffset offsetfile'");
	puts (" -l: first line to execute in scorefile (if specified) else in cutfile");
	puts (" -ox: X offset in millimeters (added to value in offsetfile)");
	puts (" -oy: Y offset in millimeters (added to value in offsetfile)");
	puts (" -oa: angle offset in radians (added to value in offsetfile)");
	puts (" -s: G code file to score before cutting");
	puts (" cutfile: G code file to cut");
}

// Same as strchr but returns "" instead of NULL if not found.
char *myStrchr (char *string, int c)
{
	char *result = strchr (string, c);

	if (result == NULL)
		return ("");
	else
		return (result);
}

// Open the filename and score it if scoring is TRUE otherwise cut.
// Return TRUE for success FALSE for error or interrupt.
BOOL DoFile (char *filename, BOOL scoring, int startLine, Prior *prior)
{
	BOOL status = TRUE;
	BOOL clockwise;
	char buffer[80];
	int totalLines, line;
	double x, y, centerX, centerY, lastEndpointX, lastEndpointY;

	FILE *infile = fopen (filename, "r");
	if (infile == NULL) {
		printf ("Cannot open file: %s\n", filename);
		return FALSE;
	}

	// Count lines
	totalLines = 0;
	while (fgets (buffer, sizeof (buffer), infile) != NULL)
		++totalLines;

	// Close and reopen the file.  Assume there will be no error.
	fclose (infile);
	infile = fopen (filename, "r");

	line = 0;
	lastEndpointX = 0.;
	lastEndpointY = 0.;
	while (1) {
		if (fgets (buffer, sizeof (buffer), infile) == NULL)
			break;
		++line;
		if (startLine != 1 && line == startLine) {
			// Reposition to where the last command would have put us.
			// Temporarily set distanceToRefocus to zero
			double saveDistanceToRefocus = prior->distanceToRefocus();
			prior->setDistanceToRefocus (0.);
			if ((status = prior->reposition (lastEndpointX, lastEndpointY)) == FALSE) {
				puts ("User aborted!");
				break;
			}
			prior->setDistanceToRefocus (saveDistanceToRefocus);

			printf ("                                                                           \r");
			printf ("\nALERT: Repositioned to line %d\n", startLine);
			printf ("  Please refocus manually.\n");
			if (prior->doInterrupt() == FALSE)
				break;
			// From now on, the commands will actually be executed.
		}

		if (line >= startLine) {
			// Only show lines starting at startLine
			printf ("                                                                           \r");
			printf ("line %d/%d: %s", line, totalLines, buffer);
		}

		if ((status = prior->checkInterrupt()) == FALSE) {
			puts ("User aborted!");
			break;
		}

		if (strnicmp (buffer, "G00 ", 4) == 0) {
			if (sscanf (myStrchr (buffer, 'X'), "X%lf", &x) != 1 ||
				sscanf (myStrchr (buffer, 'Y'), "Y%lf", &y) != 1)
				puts ("Syntax error");
			else {
				lastEndpointX = x;
				lastEndpointY = y;
				if (line < startLine)
					continue;

				if ((status = prior->reposition (x, y)) == FALSE) {
					puts ("User aborted!");
					break;
				}
			}
		} else if (strnicmp (buffer, "G01 ", 4) == 0 ||
				   strnicmp (buffer, "X", 1) == 0 ||
				   strnicmp (buffer, "Y", 1) == 0) {
			if (sscanf (myStrchr (buffer, 'X'), "X%lf", &x) != 1 ||
				sscanf (myStrchr (buffer, 'Y'), "Y%lf", &y) != 1)
				puts ("Syntax error");
			else {
				lastEndpointX = x;
				lastEndpointY = y;
				if (line < startLine)
					continue;

				if (scoring) {
					if ((status = prior->scoreTo (x, y)) == FALSE) {
						puts ("User aborted!");
						break;
					}
				} else {
					if ((status = prior->cutTo (x, y)) == FALSE) {
						puts ("User aborted!");
						break;
					}
				}
			}
		} else if (strnicmp (buffer, "G02 ", 4) == 0 ||
				   strnicmp (buffer, "G03 ", 4) == 0) {
			if (sscanf (myStrchr (buffer, 'X'), "X%lf", &x) != 1 ||
				sscanf (myStrchr (buffer, 'Y'), "Y%lf", &y) != 1 ||
				sscanf (myStrchr (buffer, 'I'), "I%lf", &centerX) != 1 ||
				sscanf (myStrchr (buffer, 'J'), "J%lf", &centerY) != 1)
				puts ("Syntax error");
			else {
				lastEndpointX = x;
				lastEndpointY = y;
				if (line < startLine)
					continue;

				clockwise = (strnicmp (buffer, "G02 ", 4) == 0) ? TRUE : FALSE;

				if (scoring) {
					if ((status = prior->scoreArcTo (x, y, centerX, centerY, clockwise)) == FALSE) {
						puts ("User aborted!");
						break;
					}
				} else {
					if ((status = prior->cutArcTo (x, y, centerX, centerY, clockwise)) == FALSE) {
						puts ("User aborted!");
						break;
					}
				}
			}
		}
	}
	
	fclose (infile);
	return status;
}

int main(int argc, char* argv[])
{
	char *cutFilename = (char *)NULL, *scoreFilename = (char *)NULL,
		*configFilename = (char *)NULL, *offsetFilename = (char *)NULL;
	BOOL scoring = FALSE;
	FILE *infile;
	int startLine = 1;
	double xOffset = 0, yOffset = 0, angleOffset = 0;

	int arg = 1;
	while (arg < argc) {
		if (argv[arg][0] != '-')
			break;

		if (strcmp (argv[arg] + 1, "c") == 0) {
			puts ("-c");
			++arg;
			if (arg >= argc) {
				Usage();
				return (0);
			}
			configFilename = argv[arg];
		}
		else if (strcmp (argv[arg] + 1, "o") == 0) {
			++arg;
			if (arg >= argc) {
				Usage();
				return (0);
			}
			offsetFilename = argv[arg];
		}
		else if (strcmp (argv[arg] + 1, "ox") == 0 ||
				 strcmp (argv[arg] + 1, "oy") == 0 ||
				 strcmp (argv[arg] + 1, "oa") == 0) {
			double value;

			++arg;
			if (arg >= argc) {
				Usage();
				return (0);
			}
			if (sscanf (argv[arg], "%lf", &value) != 1) {
				Usage();
				return (0);
			}
			
			if (strcmp (argv[arg-1] + 1, "ox") == 0)
				xOffset = value;
			if (strcmp (argv[arg-1] + 1, "oy") == 0)
				yOffset = value;
			if (strcmp (argv[arg-1] + 1, "oa") == 0)
				angleOffset = value;
		}
		else if (strcmp (argv[arg] + 1, "s") == 0) {
			++arg;
			if (arg >= argc) {
				Usage();
				return (0);
			}
			scoreFilename = argv[arg];
		}
		else if (strcmp (argv[arg] + 1, "l") == 0) {
			++arg;
			if (arg >= argc) {
				Usage();
				return (0);
			}
			if (sscanf (argv[arg], "%d", &startLine) != 1 || startLine < 1) {
				puts ("-l startline must be an integer >= 1");
				return (0);
			}
		}
		else {
			Usage();
			return (0);
		}

		++arg;
	}

	if (arg < argc) {
		// Get non-option parameter as cut filename
		cutFilename = argv[arg];
		++arg;
	}

	if (arg < argc) {
		// Had more than one non-option parameter
		Usage();
		return (0);
	}

	if (scoreFilename == NULL && cutFilename == NULL) {
		Usage ();
		return (0);
	}

	if (scoreFilename != NULL) {
		// Try to open
		infile = fopen (scoreFilename, "r");
		if (infile == NULL) {
			// Assume couldn't open because doesn't exist
			printf ("Can't open score file: %s\n", scoreFilename);
			return (0);
		}
		else
			// Only used infile to test for existence
			fclose (infile);
	}
	if (cutFilename != NULL) {
		// Try to open
		infile = fopen (cutFilename, "r");
		if (infile == NULL) {
			// Assume couldn't open because doesn't exist
			printf ("Can't open cut file: %s\n", cutFilename);
			return (0);
		}
		else
			// Only used infile to test for existence
			fclose (infile);
	}

	printf ("Using config %s, scorefile %s, cutfile %s\n",
			(configFilename == NULL) ? "<default>": configFilename,
			(scoreFilename == NULL) ? "<none>": scoreFilename,
			(cutFilename == NULL) ? "<none>": cutFilename);

	MicrolazeConfig config;
	// If no configFilename, config already has defaults
	if (configFilename != NULL) {
		if (config.loadFromFile (configFilename) == FALSE)
			// already printed error
			return (0);
	}
	
	MicrolazeOffset offset;
	// If no offsetFilename, offset already has defaults
	if (offsetFilename != NULL) {
		if (offset.loadFromFile (offsetFilename) == FALSE)
			// already printed error
			return (0);
	}
	// Add values from -ox, -oy and -oa command line options.
	offset.setXOffset(offset.xOffset() + xOffset);
	offset.setYOffset(offset.yOffset() + yOffset);
	offset.setAngleOffset(offset.angleOffset() + angleOffset);
	
	printf ("Offsets: X %.3lf mm, Y %.3lf mm, angle %.2lf radians\n",
			offset.xOffset(), offset.yOffset(), offset.angleOffset());
	printf ("Laser: %s %s %d, pulse %d Hz, markersize %dx%d\n",
			config.wavelengthString(), config.energyRangeString(), config.energyLevel(), 
			config.pulseRate(),	config.markerSizeX(), config.markerSizeY());
	printf ("Score step %.3lf mm, hold %.1lf sec. Cut step %.3lf mm, hold %.1lf sec, speed %d.\n",
		    config.scoreStepSize(), config.scoreHoldTime(),
			config.cutStepSize(), config.cutHoldTime(), config.cutFeedSpeed());
	printf ("Distance to refocus %.3lf mm, Focus offcenter %.3lf mm, Min score %d\n", config.distanceToRefocus(),
			config.focusOffcenterDistance(), config.minFocusScore());

	Quiklaze quiklaze
		(config.wavelength(), config.energyRange(), config.energyLevel(), config.pulseRate(), 
		 config.markerSizeX(), config.markerSizeY());	
	// Use ConsolePrior defined above to get interrupt from the keyboard.
	ConsolePrior prior;
	prior.setQuiklaze (&quiklaze);

	prior.setScoreStepSize (config.scoreStepSize());
	prior.setCutStepSize (config.cutStepSize());
	prior.setDistanceToRefocus (config.distanceToRefocus());
	prior.setFocusOffcenterDistance (config.focusOffcenterDistance());
	prior.setScoreHoldTime (config.scoreHoldTime());
	prior.setCutHoldTime (config.cutHoldTime());
	prior.setMinFocusScore (config.minFocusScore());
	prior.setCutFeedSpeed (config.cutFeedSpeed());

	prior.setXOffset (offset.xOffset());
	prior.setYOffset (offset.yOffset());
	prior.setAngleOffset (offset.angleOffset());

	puts ("Hit any key to pause.");

	if (scoreFilename != NULL) {
		printf ("                                                                           \r");
		printf ("\nScoring %s ...\n", scoreFilename);
		if (DoFile (scoreFilename, TRUE, startLine, &prior) == FALSE)
			// Already printed error
			return (0);

		// If there was a startline given, we already skipped to it in the score file,
		//   so the cut file starts at line 1.
		startLine = 1;
	}
	if (cutFilename != NULL) {
		printf ("                                                                           \r");
		printf ("\nCutting %s ...\n", cutFilename);
		if (DoFile (cutFilename, FALSE, startLine, &prior) == FALSE)
			// Already printed error
			return (0);
	}

	struct tm *newtime;
	time_t aclock;

	time( &aclock );                 /* Get time in seconds */
	newtime = localtime( &aclock );  /* Convert time to struct to tm form */
	// Print job finish time
	printf ("                                                                           \r");
	printf("\nJob finished at %s", asctime(newtime));

	return 0;
}
