/* No representations are made concerning either the merchantability of
   this software or the suitability of this software for any particular
   purpose. It is provided "as is" without express or implied warranty
   of any kind.

   License to copy and use this software is granted provided that these
   notices are retained in any copies of any part of this documentation
   and/or software.
 */

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <windowsx.h>
#include <process.h>
#include <math.h>
#include "serial.h"
#include "quiklaze.h"
#include "prior.h"

static const double SECONDS_PER_IO = .8;
static const double PI = 3.1415926535897932384626433832795;

// If doHome is TRUE, this will move to 0,0 in contructor and destructor
Prior::Prior (BOOL doHome) 
{
	initialize (doHome);
}

// Default to TRUE for doHome
Prior::Prior () 
{
	initialize (TRUE);
}

void Prior::initialize (BOOL doHome)
{
	_doHome = doHome;
	_quiklaze = (Quiklaze *)NULL;

	// Assume already focussed at beginning.
	_focussedX = 0.;
	_focussedY = 0.;

	_distanceToRefocus = .5;
	_focusOffcenterDistance = .1;
	_scoreStepSize = .15;
	_cutStepSize = .025;
	_scoreHoldTime = 2.;
	_cutHoldTime = 1.;
	_minFocusScore = 800;
	_cutFeedSpeed = 100;

	_xOffset = 0.;
	_yOffset = 0.;
	_angleOffset = 0.;
	_cosAngle = 1.;
	_sinAngle = 0.;

	if (OpenLine(&_hComm, "COM1", CBR_9600) == 0) {
		puts ("Prior: Error in OpenLine");
	}

	// Clear out buffers
	command ("P");

	// Force setting of speed
	_speed = 0;
	setSpeed (100);

	stopFiring();
	if (_doHome)
		// Note that _xOffset and _yOffset have not been set.
		moveTo (0.,0.);
}

Prior::~Prior ()
{
	stopFiring();

	if (_doHome) {
		// Undo x and y offsets so we go to the true home.
		_xOffset = 0.;
		_yOffset = 0.;
		setSpeed (100);
		moveTo (0., 0.);
	}

	CloseHandle (_hComm);
}

// If _quiklaze is not null, call its startFiring()
void Prior::startFiring()
{
	if (_quiklaze != (Quiklaze *)NULL)
		_quiklaze->startFiring();
}

// If _quiklaze is not null, call its stopFiring()
void Prior::stopFiring()
{
	if (_quiklaze != (Quiklaze *)NULL)
		_quiklaze->stopFiring();
}

// In mm.  If zero, it means do not refocus.
// This does nothing if value is negative.
void Prior::setDistanceToRefocus (double distanceToRefocus) 
{
	if (distanceToRefocus < 0.)
		return;

	_distanceToRefocus = distanceToRefocus;
}

// In mm.  This does nothing if value is negative
void Prior::setFocusOffcenterDistance (double focusOffcenterDistance) 
{
	if (focusOffcenterDistance < 0.)
		return;

	_focusOffcenterDistance = focusOffcenterDistance;
}

// In mm.  This does nothing if step size is negative or zero.
void Prior::setScoreStepSize (double scoreStepSize) 
{
	if (scoreStepSize <= 0.)
		return;

	_scoreStepSize = scoreStepSize;
}

// In mm.  This does nothing if step size is negative or zero.
void Prior::setCutStepSize (double cutStepSize) 
{
	if (cutStepSize <= 0.)
		return;

	_cutStepSize = cutStepSize;
}

// In mm.  This does nothing if value is negative
void Prior::setScoreHoldTime (double scoreHoldTime) 
{
	if (scoreHoldTime < 0.)
		return;

	_scoreHoldTime = scoreHoldTime;
}

// In mm.  This does nothing if value is negative
void Prior::setCutHoldTime (double cutHoldTime) 
{
	if (cutHoldTime < 0.)
		return;

	_cutHoldTime = cutHoldTime;
}

// This does nothing if value is negative
void Prior::setMinFocusScore (int minFocusScore) 
{
	if (minFocusScore < 0)
		return;

	_minFocusScore = minFocusScore;
}

// This does nothing if value is outside range 1 to 100.
void Prior::setCutFeedSpeed (int cutFeedSpeed) 
{
	if (cutFeedSpeed < 1 || cutFeedSpeed > 100)
		return;

	_cutFeedSpeed = cutFeedSpeed;
}

// In radians
void Prior::setAngleOffset (double angleOffset) 
{
	_angleOffset = angleOffset;

	// Precompute sin and cos
	_cosAngle = cos (angleOffset);
	_sinAngle = sin (angleOffset);
}

// Send the commandString
void Prior::command (char *commandString) 
{
	char buffer[80];

	command (commandString, buffer, sizeof (buffer));
}

// Send the commandString and return the result in the buffer of size bufferSize
void Prior::command (char *commandString, char *buffer, DWORD bufferSize) 
{
	printf ("                                                                           \r");
	printf ("  Prior send: %s\r", commandString);
	if (PutsRS232(_hComm, commandString) != TRUE) {
		puts ("Prior: error in PutsRS232");
		return;
	}			
	if (GetsRS232(_hComm, buffer, bufferSize) != TRUE) {
		puts ("Prior: error in GetsRS232");
		return;
	}
}

// Set the moving speed.  speed is from 1 to 100 (100 = 100% max speed).
void Prior::setSpeed (int speed)
{
	char buffer[20];

	if (speed < 1 || speed > 100)
		// erroneous value
		return;

	if (_speed == speed)
		// already at this speed.
		return;

	sprintf (buffer, "SMX,%d", speed);
	command (buffer);
	sprintf (buffer, "SRX,%d", speed);
	command (buffer);
	sprintf (buffer, "SFX,%d", speed);
	command (buffer);

	_speed = speed;
}

// x and y are in mm
// This is low-level and does not change the speed.  Merely issue G command.
void Prior::moveTo (double x, double y) 
{
	char buffer[80];
	
	double adjustedX = x * _cosAngle - y * _sinAngle + _xOffset;
	double adjustedY = x * _sinAngle + y * _cosAngle + _yOffset;

	// The Prior stage seems to reverse the X
	long int xi = -(long int)(adjustedX * STEPS_PER_MM);
	long int yi = (long int)(adjustedY * STEPS_PER_MM);

	sprintf (buffer, "G,%li,%li", xi, yi);
	command (buffer);

	_lastX = x;
	_lastY = y;
}

// x and y are in mm
// This turns off the laser and moves to x,y at maximum speed
// If _distanceToRefocuss is not zero, this focusses every _distanceToRefocuss.
// If already at x,y, this does nothing.
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::reposition (double x, double y) 
{
	if ((_lastX == x) && (_lastY == y))
		return TRUE;

    stopFiring();
	setSpeed (100);
	if (_distanceToRefocus == 0.0) {
		// Simply move without refocussing
		moveTo (x,y);
		if (checkInterrupt() == FALSE)
			return FALSE;

		return TRUE;
	}

	// We move on a straight line to the destination, jumping by _distanceToRefocus
	//   and focussing
	double deltaX = x - _lastX, deltaY = y - _lastY;
	double incrementX, incrementY;

	BOOL deltaXBigger = (fabs(deltaX) > fabs(deltaY));

	if (deltaXBigger) {
		incrementX = _distanceToRefocus * (deltaX / fabs (deltaX));
		incrementY = _distanceToRefocus * (deltaY / fabs (deltaX));
	} else {
		incrementY = _distanceToRefocus * (deltaY / fabs (deltaY));
		incrementX = _distanceToRefocus * (deltaX / fabs (deltaY));
	}

	while (1) {
		if (deltaXBigger) {
			if (fabs (x - _lastX) < _distanceToRefocus * 1.1)
				break;
		} else {
			if (fabs (y - _lastY) < _distanceToRefocus * 1.1)
				break;
		}

		// Use focusCommand() instead of focus() to avoid moving off axis.
		moveTo (_lastX + incrementX, _lastY + incrementY);
		// Use 0. for focusOffcenterDistance since we are not cutting
		if (focus (0.) == FALSE)
			return FALSE;
		if (checkInterrupt() == FALSE)
			return FALSE;
	}

	moveTo (x, y);
	if (checkInterrupt() == FALSE)
		return FALSE;

	return TRUE;
}

// x and y are in mm
// This cuts by moving in steps of _cutStepSize.  (Should be half a spot size.)
// If already at x,y, this does nothing.
// This calls checkFocus() before each move
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::cutTo (double x, double y) 
{
	return (cutToHelper (x, y, _cutStepSize, _cutHoldTime, _cutFeedSpeed));
}

// x and y are in mm
// This scores by drilling holes along the path to x,y at each scoreStepSize()
// If already at x,y, this does nothing.
// This calls checkFocus() before each move
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::scoreTo (double x, double y) 
{
	return (cutToHelper (x, y, _scoreStepSize, _scoreHoldTime, 100));
}

// x and y are in mm
// This cuts by moving in steps of stepSize, holding for holdTime at the given feed speed.
// If already at x,y, this does nothing.
// This calls checkFocus() before each move
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::cutToHelper (double x, double y, double stepSize, double holdTime, int feedSpeed) 
{
	if ((_lastX == x) && (_lastY == y))
		return TRUE;

	double deltaX = x - _lastX, deltaY = y - _lastY;
	double incrementX, incrementY;

	BOOL deltaXBigger = (fabs(deltaX) > fabs(deltaY));
	// distance is total distance of the cut
	double distance = sqrt(deltaX * deltaX + deltaY * deltaY);

	if (deltaXBigger) {
		incrementX = stepSize * (deltaX / fabs (deltaX));
		incrementY = stepSize * (deltaY / fabs (deltaX));

		// Now shrink these values proportionally so that a cut at 45 degree angle
		//    will have an increment of stepSize * cos(45), to keep the line width the same.
		incrementX *= fabs(deltaX) / distance;
		incrementY *= fabs(deltaX) / distance;
	} else {
		incrementY = stepSize * (deltaY / fabs (deltaY));
		incrementX = stepSize * (deltaX / fabs (deltaY));

		// Now shrink these values proportionally so that a cut at 45 degree angle
		//    will have an increment of stepSize * cos(45), to keep the line width the same.
		incrementX *= fabs(deltaY) / distance;
		incrementY *= fabs(deltaY) / distance;
	}

	double extraPause = holdTime - SECONDS_PER_IO;
	if (extraPause < 0.)
		extraPause = 0.;

	setSpeed (feedSpeed);

	// Fire for the hold time to get the first "fencepost"
	if (checkFocus () == FALSE)
		return FALSE;
	startFiring();
	if (checkInterrupt() == FALSE)
		return FALSE;
	// Stay here to make the hole deep enough
	PauseSeconds (extraPause);
	if (checkInterrupt() == FALSE)
		return FALSE;

	while (1) {
		if (deltaXBigger) {
			if (fabs (x - _lastX) < stepSize * 1.1)
				break;
		} else {
			if (fabs (y - _lastY) < stepSize * 1.1)
				break;
		}

		if (checkFocus () == FALSE)
			return FALSE;
		startFiring();
		moveTo (_lastX + incrementX, _lastY + incrementY);
		if (checkInterrupt() == FALSE)
			return FALSE;
		// Stay here to make the hole deep enough
		PauseSeconds (extraPause);
		if (checkInterrupt() == FALSE)
			return FALSE;
	}
 
	if (checkFocus() == FALSE)
		return FALSE;
	startFiring();
	moveTo (x, y);
	if (checkInterrupt() == FALSE)
		return FALSE;
	PauseSeconds (extraPause);
	if (checkInterrupt() == FALSE)
		return FALSE;

	return TRUE;
}

// x and y are in mm
// This cuts to x,y on an arc around centerX,centerY by moving in steps of 
//   _cutStepSize.  (Should be half a spot size.)  If clockwise is TRUE, cut
//   clockwise, else counterclockwise.
// This calls checkFocus() before each move
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::cutArcTo (double x, double y, double centerX, double centerY, BOOL clockwise) 
{
	return (cutArcToHelper
			(x, y, centerX, centerY, clockwise, _cutStepSize, _cutHoldTime, _cutFeedSpeed));
}

// x and y are in mm
// This cuts to x,y on an arc around centerX,centerY by drilling in steps of 
//   scoreStepSize().  If clockwise is TRUE, cut clockwise, else counterclockwise.
// This calls checkFocus() before each move
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::scoreArcTo (double x, double y, double centerX, double centerY, BOOL clockwise) 
{
	return (cutArcToHelper
			(x, y, centerX, centerY, clockwise, _scoreStepSize, _scoreHoldTime, 100));
}

// Query the stage and return the current x and y positions in millimeters.
void Prior::getXY (double *x, double *y)
{
	double xSteps, ySteps;
	char buffer[80];

	command ("P", buffer, sizeof (buffer));
	sscanf (buffer, "%lf", &xSteps);
	sscanf (strchr (buffer, ','), ",%lf", &ySteps);
	
	// The Prior stage seems to reverse the X
	*x = -xSteps / STEPS_PER_MM;
	*y = ySteps / STEPS_PER_MM;
}

// x and y are in mm
// This cuts to x,y on an arc around centerX,centerY by moving in steps of 
//   stepSize, holding for holdTime at the given feedSpeed. If clockwise is TRUE, cut
//   clockwise, else counterclockwise.
// This calls checkFocus() before each move
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::cutArcToHelper 
  (double x, double y, double centerX, double centerY, BOOL clockwise, double stepSize,
   double holdTime, int feedSpeed) 
{
	double radius, dxStart, dyStart, dxEnd, dyEnd, angleStart, angleEnd, angleIncrement,
		angle;
	BOOL fullCircle;

	dxStart = _lastX - centerX;
	dyStart = _lastY - centerY;
	dxEnd = x - centerX;
	dyEnd = y - centerY;
	if ((dxStart == 0 && dyStart == 0) || (dxEnd == 0 && dyEnd == 0))
		// Can't have arc with zero radius.  (don't expect this to happen.)
		return (TRUE);

	radius = sqrt (dxStart * dxStart + dyStart * dyStart);
	// compute angle of the start and end points on the circle
	angleStart = atan2 (dyStart, dxStart);
	angleEnd = atan2 (dyEnd, dxEnd);
	// The angle increment will yeild a length of stepSize along the circle
	angleIncrement = stepSize / radius;

	// If the start and endpoints are the same, we will need to make angleEnd
	//    go around a full circle.
	fullCircle = (_lastX == x && _lastY == y) ? TRUE : FALSE;
	if (clockwise) {
		// clockwise has negative angle increment
		angleIncrement = -angleIncrement;

		if (fullCircle || angleEnd > angleStart)
			angleEnd -= 2*PI;
	}
	else {
		// going counterclockwise (positige angle increment)
		if (fullCircle || angleEnd < angleStart)
			angleEnd += 2*PI;
	}

	double extraPause = holdTime - SECONDS_PER_IO;
	if (extraPause < 0.)
		extraPause = 0.;

	setSpeed (feedSpeed);

	// Fire for the hold time to get the first "fencepost"
	if (checkFocus () == FALSE)
		return FALSE;
	startFiring();
	if (checkInterrupt() == FALSE)
		return FALSE;
	// Stay here to make the hole deep enough
	PauseSeconds (extraPause);
	if (checkInterrupt() == FALSE)
		return FALSE;
	
	angle = angleStart;
	while (1) {
		if (fabs (angleEnd - angle) < fabs (angleIncrement * 1.1))
			break;
		angle += angleIncrement;

		if (checkFocus () == FALSE)
			return FALSE;
	    startFiring();
		moveTo (centerX + radius * cos (angle), centerY + radius * sin (angle));
		if (checkInterrupt() == FALSE)
			return FALSE;
		// Stay here to make the hole deep enough
		PauseSeconds (extraPause);
		if (checkInterrupt() == FALSE)
			return FALSE;
	}

	if (checkFocus() == FALSE)
		return FALSE;
	startFiring();
	moveTo (x, y);
	if (checkInterrupt() == FALSE)
		return FALSE;
	PauseSeconds (extraPause);
	if (checkInterrupt() == FALSE)
		return FALSE;

	return TRUE;
}

// This stops firing.  Calling routine must do startFiring() afterwards if necessary.
// This moves offcenter by focusOffcenterDistance before focussing.
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::focus(double focusOffcenterDistance) 
{
	char buffer[80];
	int afs;
	double saveLastX = _lastX, saveLastY = _lastY;

	// If already stopped, this does nothing
	stopFiring();

	if (focusOffcenterDistance != 0.)
		moveTo (_lastX + focusOffcenterDistance, _lastY + focusOffcenterDistance);
	if (checkInterrupt() == FALSE)
		return FALSE;

	command ("A,3");
	// Now get the autofocus score
	command ("AFS", buffer, sizeof (buffer));
	sscanf (buffer, "%d", &afs);
	if (afs < _minFocusScore) {
		printf ("                                                                           \r");
		printf ("\nWARNING: Autofocus score %d is below minimum of %d.\n", afs, _minFocusScore);
		printf ("Please refocus manually.\n");
		if (doInterrupt() == FALSE)
			return FALSE;
	}

	if (checkInterrupt() == FALSE)
		return FALSE;
	if (focusOffcenterDistance != 0.)
		moveTo (saveLastX, saveLastY);

	_focussedX = _lastX;
	_focussedY = _lastY;

	return TRUE;
}

// If _distanceToRefocus is zero, don't do anything.
// If moved since the last focus a distange greater than value set given to setDistanceToRefocus(),
//    then run focus().
// Note if focus() called, this stops firing, so calling routine must call startFiring().
// Return TRUE for success, FALSE for user abort (uses checkInterrupt()).
BOOL Prior::checkFocus () 
{
	if (_distanceToRefocus == 0.)
		return TRUE;

	double deltaX = _lastX - _focussedX;
	double deltaY = _lastY - _focussedY;

	if ((deltaX*deltaX + deltaY*deltaY) > (_distanceToRefocus * _distanceToRefocus)) {
		// do autofocus
		if (focus(_focusOffcenterDistance) == FALSE)
			return FALSE;
	}

	return TRUE;
}

// Return FALSE to abort, TRUE to continue operation.
// This should check for the interrupt event and if there is one, call doInterrupt.
// This may stop firing.  Calling routine must do startFiring() afterwards if necessary.
// This is called frequently during operation to give the user a chance to interrupt.
// (This is the default method which always returns TRUE. Can override with a subclass.)
BOOL Prior::checkInterrupt () 
{
	return (TRUE);
}

// Return FALSE to abort, TRUE to continue operation.
// This can be called by checkInterrupt or, for example, when focus is out of range.
// Should prompt the user and allow to continue or abort.
// This will stop firing.  Calling routine must do startFiring() afterwards if necessary.
// (This is the default method which always returns TRUE. Can override with a subclass.)
BOOL Prior::doInterrupt () 
{
	stopFiring();

	return (TRUE);
}
