/* No representations are made concerning either the merchantability of
   this software or the suitability of this software for any particular
   purpose. It is provided "as is" without express or implied warranty
   of any kind.

   License to copy and use this software is granted provided that these
   notices are retained in any copies of any part of this documentation
   and/or software.
 */

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <windowsx.h>
#include <process.h>
#include "serial.h"
#include "quiklaze.h"

Quiklaze::Quiklaze () 
{
	// Use defaults.
	initialize (2, 2, 1000, 40, 100, 100);
}

// See individual setWavelength, etc. for what values to use.
// It's useful to use this instead of just using the default constructor and calling
//   setWavelength, etc. because that would take extra time to talk to the serial port.
Quiklaze::Quiklaze 
  (int wavelength, int energyRange, int energyLevel, int pulseRate, 
   int markerSizeX, int markerSizeY) 
{
	initialize (wavelength, energyRange, energyLevel, pulseRate, markerSizeX, markerSizeY);
}

void Quiklaze::initialize 
  (int wavelength, int energyRange, int energyLevel, int pulseRate, 
   int markerSizeX, int markerSizeY)
{
	if (OpenLine(&_hComm, "COM2", CBR_9600) == 0) {
		puts ("Quiklaze: Error in OpenLine");
	}

	command ("P");

	// Turn on
	command ("@1");
	// Pause here because @1 seems to take a long time to execute
	PauseSeconds (2.);

	// Force sending stop firing command
	_firing = TRUE;
	stopFiring();

	// Force all these to be set
	_wavelength = -1;
	setWavelength (wavelength);
	_energyRange = -1;
	_energyLevel = -1;
	setEnergyRange (energyRange);
	setEnergyLevel (energyLevel);
	_pulseRate = -1;
	setPulseRate (pulseRate);
	_markerSizeX = -1;  
	_markerSizeY = -1;
	setMarkerSize (markerSizeX, markerSizeY);
}

Quiklaze::~Quiklaze () 
{
	stopFiring();
	command ("@0");
	// Pause here because @1 seems to take a long time to execute
	PauseSeconds (1.);
	CloseHandle (_hComm);
}

// Send the command with a '\015'.
// However, if the command is "P", don't send the '\015'.
void Quiklaze::command (char *command) {
		char buffer[80];

		printf ("                                                                           \r");
		printf ("  Qlaze send: %s\r", command);
		if (strcmp (command, "P") == 0) {
			if (WriteToRS232(_hComm, (unsigned char *)"P", 1) != TRUE) {
				puts ("Quiklaze: error in WriteToRS232");
				return;
			}
		} else {
			if (PutsRS232(_hComm, command) != TRUE) {
				puts ("Quiklaze: error in PutsRS232");
				return;
			}
		}

		// Allow time for response.  Without this pause, the input buffer from
		//   the Quiklaze doesn't seem to fill in time.
		PauseSeconds (.5);

		// It seems Quiklaze returns LFCR so we can use GetRS232 which looks for CR
		if (GetsRS232(_hComm, buffer, sizeof (buffer)) != TRUE) {
			puts ("Quiklaze: error in GetsRS232");
			return;
		}
}

// wavelength is 1 for IR, 2 for Green, 3 for UV
// If wavelength already has the value or is out of this range, this does not change it.
void Quiklaze::setWavelength (int wavelength)
{
	char buffer[20];

	if (_wavelength != wavelength && wavelength >= 1 && wavelength <= 3) {
		sprintf (buffer, "8%d", wavelength);
		command (buffer);
		_wavelength = wavelength;
	}
}

// range is 1 for low, 2 for high
// If range already has the value or is invalid, this does not change it.
void Quiklaze::setEnergyRange (int range)
{
	char buffer[20];

	if (_energyRange != range && range >= 1 && range <= 2) {
		sprintf (buffer, "4%d", range);
		command (buffer);
		_energyRange = range;
	}
}

// level is from 0 to 1000
// If level already has the value or is out of range, this does not change it.
void Quiklaze::setEnergyLevel (int level)
{
	char buffer[20];

	if (_energyLevel != level && level >= 0 && level <= 1000) {
		// Quiklaze takes value from 0 to 255
		sprintf (buffer, "1%d", (level * 255) / 1000);
		command (buffer);
		_energyLevel = level;
	}
}

// pulseRate is from 1 to 40
// If pulseRate already has the value or is out of range, this does not change it.
void Quiklaze::setPulseRate (int pulseRate)
{
	char buffer[20];

	if (_pulseRate != pulseRate && pulseRate >= 1 && pulseRate <= 40) {
		sprintf (buffer, "A%d", pulseRate);
		command (buffer);
		_pulseRate = pulseRate;
	}
}

// x and y are from 0 to 100.
// If x or y is out of this range, this does not change it.
// If x or y is already the given value, this doesn't change it.
void Quiklaze::setMarkerSize (int x, int y)
{
	char buffer[20];

	if (_markerSizeX != x && x >= 0 && x <= 100) {
		// The Quiklaze takes a range from 0 to 255
		sprintf (buffer, "2%d", (x * 255) / 100);
		command (buffer);
		_markerSizeX = x;
	}
	if (_markerSizeY != y && y >= 0 && y <= 100) {
		// The Quiklaze takes a range from 0 to 255
		sprintf (buffer, "3%d", (y * 255) / 100);
		command (buffer);
		_markerSizeY = y;
	}
}

// If not already firing, start firing.
void Quiklaze::startFiring () 
{
	if (_firing)
		// already firing
		return;

	command ("51");
	command ("7");
	// Seems to take a moment to start firing
	PauseSeconds (1.);
	_firing = TRUE;
}

// If not already stopped, stop firing
void Quiklaze::stopFiring () 
{
	if (!_firing)
		// already stopped
		return;

	command ("50");
	_firing = FALSE;
}
